import { CollectionViewer, DataSource, SelectionChange } from "@angular/cdk/collections";
import { FlatTreeControl } from "@angular/cdk/tree";
import { BehaviorSubject, merge, Observable } from "rxjs";
import { map } from "rxjs/operators";
import { DynamicDatabase } from "./dynamic-database";
import { RuleFlatNode } from "./rule-flat-node";

export class DynamicDataSource implements DataSource<RuleFlatNode> {

  dataChange = new BehaviorSubject<RuleFlatNode[]>([]);

  get data(): RuleFlatNode[] { return this.dataChange.value; }
  set data(value: RuleFlatNode[]) {
    this._treeControl.dataNodes = value;
    this.dataChange.next(value);
  }

  constructor(private _treeControl: FlatTreeControl<RuleFlatNode>,
    private _database: DynamicDatabase) {
  
     }

  connect(collectionViewer: CollectionViewer): Observable<RuleFlatNode[]> {
    this._treeControl.expansionModel.changed.subscribe(change => {
      if ((change as SelectionChange<RuleFlatNode>).added ||
        (change as SelectionChange<RuleFlatNode>).removed) {
        this.handleTreeControl(change as SelectionChange<RuleFlatNode>);
      }
    });

    return merge(collectionViewer.viewChange, this.dataChange).pipe(map(() => this.data));
  }

  disconnect(collectionViewer: CollectionViewer): void { }

  /** Handle expand/collapse behaviors */
  handleTreeControl(change: SelectionChange<RuleFlatNode>) {
    if (change.added) {
      change.added.forEach(node => this.toggleNode(node, true));
    }
    if (change.removed) {
      change.removed.slice().reverse().forEach(node => this.toggleNode(node, false));
    }
  }

  /**
   * Toggle the node, remove from display list
   */
  toggleNode(node: RuleFlatNode, expand: boolean) {
    const children = node.item.subRule;
    const index = this.data.indexOf(node);
    if (!children || index < 0) { // If no children, or cannot find the node, no op
      return;
    }

    setTimeout(() => {
      if (expand) {
        const nodes = children.map(name =>
          new RuleFlatNode(
            name,
            node.level + 1,
            !!node.item.subRule ? node.item.subRule.length > 0 : false)
        );
        this.data.splice(index + 1, 0, ...nodes);
      } else {
        let count = 0;
        for (let i = index + 1; i < this.data.length
          && this.data[i].level > node.level; i++, count++) { }
        this.data.splice(index + 1, count);
      }

      // notify the change
      this.dataChange.next(this.data);
      node.isLoading = false;
    }, 1000);
  }
}