import { Injectable } from "@angular/core";
import { BehaviorSubject, Subject } from "rxjs";
import { CreateRuleDto, Rule, RulesService } from "../../../../../api";
import { environment } from "../../../../environments/environment";
import { RuleDataService } from "../../../service/rule-data.service";
import { EventMessage } from "./event-message";


@Injectable({ providedIn: 'root' })
export class DynamicDatabase {

  dataChange = new BehaviorSubject<Rule[]>([]);
  childAdded = new Subject<Rule>();
  onEvent = new Subject<EventMessage>()

  get data(): Rule[] { return this.dataChange.value; }
  get count(): number { return this.dataChange.value.length };
  protected removeData(rule: Rule) {
    const indexOfRule = this.dataChange.value.indexOf(rule);
    if (indexOfRule !== -1) {
      this.dataChange.value.splice(indexOfRule, 1)
      this.dataChange.next(
        this.data
      );
    }
  }
  protected addData(rule: Rule) { this.dataChange.next([...this.dataChange.value, rule]) }
  constructor(protected rulesService: RulesService, protected ruleHelperService: RuleDataService) {
    this.rulesService.configuration.basePath = environment.apiUrl;
    this.rulesService.findAllAdmin()
      .subscribe(
        (rules: Rule[]) => this.dataChange.next(
          rules
            .filter(ruleHelperService.filterRuleWithParentFromRootLayer)
            .sort(ruleHelperService.compareRulesByParagraph)
        )
      ),
      (err) => {
        this.onEvent.next(
          {
            event: 'Error',
            error: err,
            message: `Fehler beim abrufen der Regeln`
          }
        );
      }
  }

  getById = (ruleId: string) => {
    const parentNode = this.getParentById(ruleId);
    if (parentNode) {
      return parentNode;
    }
    for (const rootRule of this.data) {
      const found = this.findChildById(rootRule, ruleId);
      if (found) {
        return found;
      }
    }
    return null;
  }

  findChildById = (parent: Rule, ruleId: string) => {
    if (!parent.subRule) {
      return null;
    }
    return parent.subRule.find((
      rule => rule.id === ruleId
    ))
  }
  getParentById = (ruleId: string) => this.data.find(rule => rule.id === ruleId)
  insertEmptyChild(parent: Rule) {
    ;
    if (!parent.subRule) {
      parent.subRule = [];
    }
    parent.subRule.push(
      {
        text: '',
        id: '0',
        parentRule: {
          id: parent.id
        },
        paragraph: this.ruleHelperService.getParagraphOfNextChild(parent),
        created: 'today',
        deletedAt: null,
        updated: null,
      } as Rule
    )
    this.dataChange.next(this.data);
    this.childAdded.next(parent);
  }

  insertEmptyParent() {
    this.addData({
      text: '',
      id: '0',
      parentRule: null,
      paragraph: `${this.count + 1}`,
      created: 'today',
      deletedAt: null,
      updated: null,
    })
  }
  insertChild(parent: Rule, subRule: Rule) {
    const createSubRule: CreateRuleDto = {
      paragraph: subRule.paragraph,
      text: subRule.text,
      subRuleIds: null,
      parentId: parent.id
    }
    this.rulesService.create(createSubRule).subscribe(
      newSubRule => {
        subRule.id = newSubRule.id;
        this.dataChange.next(this.data);
        if (typeof newSubRule.id !== 'undefined' && typeof newSubRule.parentRule?.id !== 'undefined') {
          const updateParentRule: CreateRuleDto = {
            paragraph: parent.paragraph,
            text: parent.text,
            subRuleIds: [...parent.subRule ? parent.subRule.map(rule => rule.id) : [], newSubRule.id]
          } as CreateRuleDto;
          this.rulesService.update(parent.id, updateParentRule).subscribe((updated) => {
            //TODO Push updated message
            this.update(parent.id, {
              paragraph: parent.paragraph,
              subRule: [
                ...parent.subRule
                  .filter(rule => rule.id !== "0"),
                subRule,
              ],
              text: parent.text
            });
          },
            (err) => {
              this.onEvent.next(
                {
                  event: 'Error',
                  error: err,
                  message: `Fehler beim aktualisieren von Regel ${parent.paragraph}`
                }
              );
            }
          )
          return;
        }
      }
    )
    if (!parent.subRule) {
      parent.subRule = [];
    }
    const lastIndex = parent.subRule.indexOf(subRule);
    parent.subRule[lastIndex] = subRule;
    this.dataChange.next(this.data);
    this.childAdded.next(parent);
  }

  reload() {
    this.dataChange.next(this.data);
  }

  update(id: string, changedData: Partial<Rule>) {
    let ruleToUpdate = this.getById(id);
    const updateDto = {
      paragraph: !!changedData.paragraph ? changedData.paragraph : ruleToUpdate.paragraph,
      text: !!changedData.text ? changedData.text : ruleToUpdate.text,
      parentId: !!changedData.parentRule ? changedData.parentRule.id : null,
      subRuleIds: [
        ...!!changedData.subRule ?
          changedData.subRule
            .map(rule => rule.id)
          : []
      ]
    } as CreateRuleDto
    this.rulesService.update(id, updateDto).subscribe(
      () => {
        this.onEvent.next(
          {
            event: 'Update',
            message: `Regel ${ruleToUpdate.paragraph} wurde erfolgreich aktualisiert`
          }
        );
      },
      (err) => {
        this.onEvent.next(
          {
            event: 'Error',
            error: err,
            message: `Fehler beim aktualisieren der Regel ${ruleToUpdate.paragraph}`
          }
        );
      }
    )
    this.dataChange.next(this.data);
  }

  insertParent(newRule: Rule) {
    const createNewRule: CreateRuleDto = {
      paragraph: newRule.paragraph,
      text: newRule.text,
      subRuleIds: null,
      parentId: null
    }
    this.rulesService.create(createNewRule).subscribe(
      (created) => {
        newRule.id = created.id;
        this.dataChange.next(this.data);
        this.onEvent.next(
          {
            event: 'Create',
            message: `Regel ${createNewRule.paragraph} wurde erfolgreich erstellt`
          }
        );
      },
      (err) => {
        this.onEvent.next(
          {
            event: 'Error',
            error: err,
            message: `Fehler beim anlegen der Regel ${createNewRule.paragraph}`
          }
        );
      }
    )
    this.dataChange.next(this.data);
  }

  removeParent(parent: Rule) {
    parent.deletedAt = "deleted";
    this.rulesService.remove(parent.id).subscribe(
      (deleted) => {
        this.onEvent.next(
          {
            event: 'Restore',
            message: `Regel ${parent.paragraph} wurde erfolgreich deaktiviert`
          }
        );
      },
      (err) => {
        this.onEvent.next(
          {
            event: 'Error',
            error: err,
            message: `Fehler beim aktivieren von Regel ${parent.paragraph}`
          }
        );
      }
    );
    this.dataChange.next(this.data);
  }

  restore(rule: Rule) {
    rule.deletedAt = null;
    if (rule.subRule) {
      rule.subRule.forEach((subRule) => {
        this.restore(subRule);
      })
    }
    this.rulesService.restore(rule.id).subscribe(
      (deleted) => {
        this.onEvent.next(
          {
            event: 'Restore',
            message: `Regel ${rule.paragraph} wurde erfolgreich reaktiviert`
          }
        );
      },
      (err) => {
        this.onEvent.next(
          {
            event: 'Error',
            error: err,
            message: `Fehler beim aktivieren der Regel ${rule.paragraph}`
          }
        );
      }
    );
    this.dataChange.next(this.data);
  }

  removeChild(parent: Rule, subRule: Rule) {
    if (parent.subRule) {
      subRule.deletedAt = "delted";
      this.rulesService.remove(subRule.id).subscribe(
        (deleted) => {
          this.update(parent.id, {
            paragraph: parent.paragraph,
            subRule: [
              ...parent.subRule
                .filter(rule => rule.id !== "0")
            ],
            text: parent.text
          });
          this.onEvent.next(
            {
              event: 'Remove',
              message: `Regel ${subRule.paragraph} wurde erfolgreich deaktiviert`
            }
          );
        },
        (err) => {
          this.onEvent.next(
            {
              event: 'Error',
              error: err,
              message: `Fehler beim  deaktivieren von Regel ${subRule.paragraph}`
            }
          );
        }
      );
      this.dataChange.next(this.data);
    }
  }

  findSubRuleFromParent = (parent: Rule, subRule: Rule): Rule | null =>
    parent.subRule.find(rules => rules.paragraph === subRule.paragraph && rules.id === subRule.id);

}