import { FlatTreeControl } from '@angular/cdk/tree';
import { Component, OnInit, TemplateRef, ViewChild } from '@angular/core';
import { FormControl } from '@angular/forms';
import { MatBottomSheet, MatBottomSheetConfig } from '@angular/material/bottom-sheet';
import { MatSnackBar } from '@angular/material/snack-bar';
import { MatTreeFlatDataSource, MatTreeFlattener } from '@angular/material/tree';
import { Subject } from 'rxjs';
import { Rule } from '../../../../api/model/rule';
import { EventDialogComponent } from '../../event-dialog/event-dialog.component';
import { OptionsShetComponent } from '../options-shet/options-shet.component';
import { DynamicDatabase } from './class/dynamic-database';
import { EventMessage } from './class/event-message';
import { RuleFlatNode } from './class/rule-flat-node';


@Component({
  selector: 'app-create-rule',
  templateUrl: './create-rule.component.html',
  styleUrls: ['./create-rule.component.scss']
})
export class CreateRuleComponent implements OnInit {

  flatNodeMap = new Map<RuleFlatNode, Rule>();
  nestedNodeMap = new Map<Rule, RuleFlatNode>();
  text = new FormControl('');
  @ViewChild(TemplateRef) template: TemplateRef<any>;


  private _transformer = (rule: Rule, level: number) => {
    ;
    const existingNode = this.nestedNodeMap.get(rule);
    const flatNode = existingNode && existingNode.item.id === rule.id
      ? existingNode
      : new RuleFlatNode(rule, level, !!rule.subRule?.length, false, level === 0)
    this.flatNodeMap.set(flatNode, rule);
    // dirty hack to expand node with new child node
    const lastChildNode = !!rule.subRule ? rule.subRule[rule.subRule.length - 1] : undefined;
    if (typeof lastChildNode !== 'undefined') {
      if (flatNode.expandable) {
        this.treeControl.expand(flatNode);
      }
    }
    return flatNode;
  }

  treeControl: FlatTreeControl<RuleFlatNode>;
  treeFlattener: MatTreeFlattener<Rule, RuleFlatNode>;
  dataSource: MatTreeFlatDataSource<Rule, RuleFlatNode>;
  _database: DynamicDatabase;

  constructor(database: DynamicDatabase, private _snackBar: MatSnackBar, readonly bottomSheet: MatBottomSheet) {
    this._database = database;
    this.treeControl = new FlatTreeControl<RuleFlatNode>(this.getLevel, this.isExpandable);
    this.treeFlattener = new MatTreeFlattener<Rule, RuleFlatNode, RuleFlatNode>(
      this._transformer, this.getLevel,
      this.isExpandable, this.getChildren
    );
    this.dataSource = new MatTreeFlatDataSource(this.treeControl, this.treeFlattener);
    this._database.dataChange.subscribe(data => {
      this.dataSource.data = data;
    });
    this._database.onEvent.subscribe(
      (event) => {
        if (event.error) {
          console.error(event.error);
        }
        this.openSnackBar(event)
      }
    )
    this._database.childAdded.subscribe(parent => this.treeControl.expand(this.nestedNodeMap.get(parent)))
  }

  ngOnInit(): void {
  }

  openOptions(node: RuleFlatNode) {
    const onRemove = new Subject<RuleFlatNode>();
    const onAdd = new Subject<RuleFlatNode>();
    const onEdit = new Subject<RuleFlatNode>();
    this.bottomSheet.open(OptionsShetComponent, {
      data: {
        remove: onRemove,
        add: onAdd,
        edit: onEdit,
        node: node
      }
    });
    onRemove.subscribe(
      (node) => {
        this.remove(node);
        this.bottomSheet.dismiss();
      }
    )
    onAdd.subscribe(
      (node) => {
        this.addEmptyRule(node);
        this.bottomSheet.dismiss();
      }
    )
    onEdit.subscribe(
      (node) => {
        this.openEdit(node);
        this.bottomSheet.dismiss();
      }
    )
  }

  openSnackBar(event: EventMessage) {
    this._snackBar.openFromComponent(EventDialogComponent, {
      data: event,
      duration: 1000,
    });
  }

  getLevel = (node: RuleFlatNode) => node.level;
  isExpandable = (node: RuleFlatNode) => node.expandable;
  getChildren = (node: Rule): Rule[] => node.subRule;
  hasChild = (_: number, node: RuleFlatNode) => node.expandable && !this.isDeleted(_, node) && !this.inEditMode(_, node);
  inEditMode = (_: number, node: RuleFlatNode) => node.editMode;
  isDeleted = (_: number, _nodeData: RuleFlatNode) => _nodeData.item.deletedAt !== null;
  hasNoContent = (_: number, _nodeData: RuleFlatNode) => _nodeData.item.text === '';
  hasContent = (_nodeData: RuleFlatNode) => _nodeData.item.text !== "";


  addEmptyRule(parentRuleNode?: RuleFlatNode) {
    const parentRule = this.flatNodeMap.get(parentRuleNode);
    if (typeof parentRule === 'undefined') {
      this._database.insertEmptyParent();
      return;
    }
    this._database.insertEmptyChild(parentRule);
  }

  getParentNode(node: RuleFlatNode): RuleFlatNode | null {
    const currentLevel = this.getLevel(node);

    if (currentLevel < 1) {
      return null;
    }

    const startIndex = this.treeControl.dataNodes.indexOf(node) - 1;

    for (let i = startIndex; i >= 0; i--) {
      const currentNode = this.treeControl.dataNodes[i];

      if (this.getLevel(currentNode) < currentLevel) {
        return currentNode;
      }
    }
    return null;
  }

  async addNode(newNode: RuleFlatNode, text: string) {
    ;
    const newRule = this.flatNodeMap.get(newNode);
    newRule.text = text;
    if (!!newRule.parentRule) {
      ;
      const parentRule = this._database.getParentById(newRule.parentRule.id)
      this._database.insertChild(parentRule, newRule)
      return;
    }
    this._database.insertParent(newRule)
  }

  async remove(nodeToRemove: RuleFlatNode) {
    const ruleToRemove = this.flatNodeMap.get(nodeToRemove);
    if (nodeToRemove.root) {
      this._database.removeParent(ruleToRemove);
      return
    }
    const parentNode = this.getParentNode(nodeToRemove);
    const parent = this.flatNodeMap.get(parentNode);
    this._database.removeChild(parent, ruleToRemove);
  }

  async openEdit(nodeToEdit: RuleFlatNode) {
    nodeToEdit.editMode = true;
    this._database.reload;
  }


  async restore(nodeToRestore: RuleFlatNode) {
    const ruleToRestore = this.flatNodeMap.get(nodeToRestore);
    this._database.restore(ruleToRestore);
  }
  async updateNode(updatedNode: RuleFlatNode, updatedText: string) {
    const parentNode = this.getParentNode(updatedNode);
    if (parentNode) {
      updatedNode.item.parentRule = {
        id: parentNode.item.id,
        text: parentNode.item.text,
        paragraph: parentNode.item.text
      }
    }
    const updatedRule = this.flatNodeMap.get(updatedNode);
    updatedRule.text = updatedText;
    this._database.update(updatedRule.id, updatedRule)
  }


}
